/*
 * mod_spin.c
 *
 * Copyright (C) 2003 - 2007 Bojan Smojver, Rexursive
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "private.h"

#include <http_config.h>
#include <http_connection.h>
#include <http_protocol.h>
#include <http_log.h>
#include <util_filter.h>

module AP_MODULE_DECLARE_DATA spin_module;

#define SPIN_MAGIC_TYPE "httpd/spin-template"
#define ERROR_BUF_SIZE  256

/* Process specific pool */
static apr_pool_t *ppool;

#if APR_HAS_THREADS
static apr_threadkey_t    *tkey;  /* thread private key */
static apr_thread_mutex_t *mutex; /* thread mutex */
#endif

/* Private data for non-threaded MPM */
static rxv_spin_private_t *priv;

/* Crypto items for session */
static char *salt,*oslt;
static ap_regex_t *ureg;

/* Maximum number of connections in the connection pool */
#define CONN_MAX 5
static unsigned conn_max=CONN_MAX;

/* Connection timeout */
#define CONN_TIMEOUT (60*1000000)

/*
 * Little helper functions...
 */

/* context creation utility */
static rxv_spin_ctx_t *context_create(apr_pool_t *pool){
  rxv_spin_ctx_t *ctx=apr_pcalloc(pool,sizeof(*ctx));

  ctx->pool=pool;
  ctx->ppool=ppool;

  ctx->data=apr_pcalloc(pool,sizeof(*ctx->data));

  ctx->data->size=1;
  ctx->data->type=RXV_SPIN_DATA_RWS;

  ctx->data->cols=apr_hash_make(pool);

  ctx->closed=RXV_SPIN_TRUE;

  return ctx;
}

/* scanner extra data creation/destruction */
static apr_status_t extra_destroy(void *data){
  rxv_spin_extra_t *extra=data;

  /* cleanup all leftover brigades */
  apr_brigade_cleanup(extra->output);
  apr_brigade_cleanup(extra->text);

  if(extra->input) apr_brigade_cleanup(extra->input);
  if(extra->split) apr_brigade_cleanup(extra->split);
  if(extra->metabb) apr_brigade_cleanup(extra->metabb);

  return APR_SUCCESS;
}

static rxv_spin_extra_t *extra_create(request_rec *r,rxv_spin_ctx_t *ctx){
  apr_pool_t *pool;
  rxv_spin_extra_t *extra=apr_pcalloc(r->pool,sizeof(*extra));

  apr_pool_create(&pool,r->pool);

  extra->pool=pool;

  extra->tail=apr_pcalloc(r->pool,sizeof(*extra->tail)*RXV_SPIN_MAX_DEPTH*2);
  extra->loop=apr_pcalloc(r->pool,sizeof(*extra->loop)*RXV_SPIN_MAX_DEPTH);
  extra->nest=apr_pcalloc(r->pool,sizeof(*extra->nest)*RXV_SPIN_MAX_DEPTH);
  extra->boff=apr_pcalloc(r->pool,sizeof(*extra->boff)*RXV_SPIN_MAX_DEPTH);

  extra->row=1; extra->col=0;
  extra->erow=extra->ecol=-1;
  extra->buf=apr_pcalloc(r->pool,RXV_SPIN_SCAN_MAX+2);

  extra->ctx=ctx?ctx:context_create(r->pool);

  extra->output=apr_brigade_create(r->pool,r->connection->bucket_alloc);
  extra->text=apr_brigade_create(r->pool,r->connection->bucket_alloc);

  apr_pool_cleanup_register(r->pool,extra,extra_destroy,apr_pool_cleanup_null);

  return extra;
}

/* private thread data creation and cleanup */
static apr_status_t private_cleanup(void *data){
  rxv_spin_private_t *priv=data;
  rxv_spin_db_t *t,*p=APR_RING_LAST(&priv->list);

  /* clean over limit and timed out connections */
  while(p!=APR_RING_SENTINEL(&priv->list,rxv_spin_db,link)){
    t=APR_RING_PREV(p,link); /* save previous - we may destroy the pool */
    if(priv->count>conn_max || p->atime+CONN_TIMEOUT<priv->atime){
      apr_pool_destroy(p->pool);
      priv->count--;
    }
    p=t;
  }

  return APR_SUCCESS;
}

static rxv_spin_private_t *private_create(){
  rxv_spin_private_t *private;
  apr_pool_t *tpool;

#if APR_HAS_THREADS
  if(mutex){
    apr_thread_mutex_lock(mutex);
    apr_pool_create(&tpool,ppool);
    apr_thread_mutex_unlock(mutex);
  } else{
    tpool=ppool;
  }
#else
  tpool=ppool;
#endif

  private=apr_pcalloc(tpool,sizeof(*private));
  private->dso=apr_hash_make(tpool);
  private->conns=apr_hash_make(tpool);
  APR_RING_INIT(&private->list,rxv_spin_db,link);
  private->pool=tpool;

  return private;
}

/*
 * Configuration...
 */
static void *create_config(apr_pool_t *p){
  rxv_spin_config_t *conf=apr_pcalloc(p,sizeof(*conf));

  conf->appinit="rxv_spin_init";
  conf->appfixup="rxv_spin_prepare";
  conf->appentry="rxv_spin_service";
  conf->storetbl="spinstore";
  conf->connpool=1;

  return conf;
}

static void *create_config_server(apr_pool_t *p,server_rec *s){
  return create_config(p);
}

static void *create_config_directory(apr_pool_t *p,char *dummy){
  return create_config(p);
}

#define merge_var(var,flag) (new->set&(flag)?new->var:old->var)

static void *merge_config(apr_pool_t *p,void *o,void *n){
  rxv_spin_config_t *old=(rxv_spin_config_t *)o,*new=(rxv_spin_config_t *)n,
                    *conf=apr_palloc(p,sizeof(*conf));

  conf->application=merge_var(application,RXV_SPIN_APPLICATION_SET);
  conf->appinit=merge_var(appinit,RXV_SPIN_APPINIT_SET);
  conf->appfixup=merge_var(appfixup,RXV_SPIN_APPFIXUP_SET);
  conf->appentry=merge_var(appentry,RXV_SPIN_APPENTRY_SET);
  conf->cfgfn=merge_var(cfgfn,RXV_SPIN_CFGFN_SET);
  conf->driver=merge_var(driver,RXV_SPIN_STORE_SET);
  conf->store=merge_var(store,RXV_SPIN_STORE_SET);
  conf->appfn=merge_var(appfn,RXV_SPIN_STORE_SET);
  conf->storetbl=merge_var(storetbl,RXV_SPIN_STORETBL_SET);
  conf->cookie=merge_var(cookie,RXV_SPIN_COOKIE_SET);
  conf->timeout=merge_var(timeout,RXV_SPIN_TIMEOUT_SET);
  conf->connpool=merge_var(connpool,RXV_SPIN_CONNPOOL_SET);
  conf->set=new->set;

  return conf;
}

static const char *set_application(cmd_parms *cmd,void *data,const char *arg){
  rxv_spin_config_t *conf=data;

  if(!(conf->application=ap_server_root_relative(cmd->pool,arg)))
    return apr_pstrcat(cmd->pool,"Invalid SpinApplication path: ",arg,NULL);

  conf->set|=RXV_SPIN_APPLICATION_SET;

  return NULL;
}

static const char *set_appinit(cmd_parms *cmd,void *data,const char *arg){
  rxv_spin_config_t *conf=data;

  conf->appinit=apr_pstrdup(cmd->pool,arg);
  conf->set|=RXV_SPIN_APPINIT_SET;

  return NULL;
}

static const char *set_appfixup(cmd_parms *cmd,void *data,const char *arg){
  rxv_spin_config_t *conf=data;

  conf->appfixup=apr_pstrdup(cmd->pool,arg);
  conf->set|=RXV_SPIN_APPFIXUP_SET;

  return NULL;
}

static const char *set_appentry(cmd_parms *cmd,void *data,const char *arg){
  rxv_spin_config_t *conf=data;

  conf->appentry=apr_pstrdup(cmd->pool,arg);
  conf->set|=RXV_SPIN_APPENTRY_SET;

  return NULL;
}

static const char *set_appconfig(cmd_parms *cmd,void *data,const char *arg){
  rxv_spin_config_t *conf=data;

  if(!(conf->cfgfn=ap_server_root_relative(cmd->pool,arg)))
    return apr_pstrcat(cmd->pool,"Invalid SpinAppConfig path: ",arg,NULL);

  conf->set|=RXV_SPIN_CFGFN_SET;

  return NULL;
}

static const char *set_store(cmd_parms *cmd,void *data,const char *arg){
  char *p;
  rxv_spin_config_t *conf=data;

  /* check if our store is file */
  if(!((p=strchr(arg,':')) && *(p+1)))
    return apr_pstrcat(cmd->pool,"Invalid SpinStore: ",arg,NULL);

  conf->driver=apr_pstrmemdup(cmd->pool,arg,p-arg);

  /* database backend */
  if(strcmp(conf->driver,"file")){
    conf->store=apr_pstrdup(cmd->pool,arg);
  } else{ /* file backend */
    if(!(conf->store=ap_server_root_relative(cmd->pool,p+1)))
      return apr_pstrcat(cmd->pool,"Invalid SpinStore: ",arg,NULL);
    else if(apr_filepath_merge(&conf->appfn,conf->store,"__application",
                               APR_FILEPATH_SECUREROOT,cmd->pool)!=APR_SUCCESS)
      return apr_pstrcat(cmd->pool,"Invalid SpinStore: ",arg,NULL);
  }

  conf->set|=RXV_SPIN_STORE_SET;

  return NULL;
}

static const char *set_storetbl(cmd_parms *cmd,void *data,const char *arg){
  rxv_spin_config_t *conf=data;

  conf->storetbl=apr_pstrdup(cmd->pool,arg);
  conf->set|=RXV_SPIN_STORETBL_SET;

  return NULL;
}

static const char *set_cookie(cmd_parms *cmd,void *data,const char *arg){
  rxv_spin_config_t *conf=data;

  conf->cookie=apr_pstrdup(cmd->pool,arg);
  conf->set|=RXV_SPIN_COOKIE_SET;

  return NULL;
}

static const char *set_timeout(cmd_parms *cmd,void *data,const char *arg){
  rxv_spin_config_t *conf=data;

  if((conf->timeout=apr_atoi64(arg)*1000000)<0)
    conf->timeout=0;

  conf->set|=RXV_SPIN_TIMEOUT_SET;

  return NULL;
}

static const char *set_connpool(cmd_parms *cmd, void *data,const char *arg){
  rxv_spin_config_t *conf=data;

  if(!strcasecmp(arg,"on"))
    conf->connpool=1;
  else if(!strcasecmp(arg,"off"))
    conf->connpool=0;
  else
    return apr_pstrcat(cmd->pool,"Invalid SpinConnPool option",arg,NULL);

  conf->set|=RXV_SPIN_CONNPOOL_SET;

  return NULL;
}

static const char *set_conncount(cmd_parms *cmd, void *data,const char *arg){
  const char *err;
  int tmp;

  if((err=ap_check_cmd_context(cmd,GLOBAL_ONLY)))
    return err;

  tmp=atoi(arg);
  conn_max=tmp<=0?CONN_MAX:tmp;

  return NULL;
}

static const char *set_salt(cmd_parms *cmd, void *data,const char *arg){
  const char *err;
  size_t len;

  if((err=ap_check_cmd_context(cmd,GLOBAL_ONLY)))
    return err;

  /* nothing to do, we've been here twice already */
  if(oslt)
    return NULL;

  if((len=strlen(arg))<RXV_SPIN_SALT_MIN)
    return apr_pstrcat(cmd->pool,"SpinSalt too short (min 30 char): ",arg,NULL);

  /* we've already seen salt once, so this must be the old one */
  if(salt){
    oslt=apr_pcalloc(cmd->pool,RXV_SPIN_SALT_MAX+1);
    memcpy(oslt,arg,len>RXV_SPIN_SALT_MAX?RXV_SPIN_SALT_MAX:len);
  } else{
    salt=apr_pcalloc(cmd->pool,RXV_SPIN_SALT_MAX+1);
    memcpy(salt,arg,len>RXV_SPIN_SALT_MAX?RXV_SPIN_SALT_MAX:len);
  }

  return NULL;
}

/*
 * Session validation functions
 */
static apr_status_t sesck_valid(apr_pool_t *pool,const char *uniq,char *hash){
  char *ndig;

  /* make HMAC from unique ID */
  if(!(ndig=rxv_spin_hmac(pool,uniq,salt)))
    return APR_EGENERAL;

  if(strcmp(ndig,hash)){
    char *odig;

    /* new salt didn't work, try the old one */
    if(!(oslt && (odig=rxv_spin_hmac(pool,uniq,oslt))) || strcmp(odig,hash))
      return APR_EGENERAL;

    /* old salt OK, but we need to replace the hash with the new one */
    memcpy(hash,ndig,APR_MD5_DIGESTSIZE*2);
  }

  return APR_SUCCESS;
}

/*
 * The other hooks...
 */
static int post_config(apr_pool_t *pconf,apr_pool_t *plog,
                       apr_pool_t *ptemp,server_rec *s){
  ap_add_version_component(pconf,RXV_SPIN_PACKAGE_SIGNATURE);
  return OK;
}

static void private_destroy(void *data){
  apr_pool_destroy(((rxv_spin_private_t *)data)->pool);
}

static void child_init(apr_pool_t *p,server_rec *s){
  int rc;
  apr_status_t rv;

  /* remember process specific pool */
  ppool=p;

#if APR_HAS_THREADS
  /* see if MPM is thread capable */
  if(ap_mpm_query(AP_MPMQ_IS_THREADED,&rc)!=APR_SUCCESS){
    ap_log_error(APLOG_MARK,APLOG_CRIT,rv,s,
                 "mod_spin: cannot get MPM threading capability");
    return;
  }

  if(rc!=AP_MPMQ_NOT_SUPPORTED){
    /* create thread private key to identify thread private data */
    if((rv=apr_threadkey_private_create(&tkey,
                                        private_destroy,ppool))!=APR_SUCCESS){
      ap_log_error(APLOG_MARK,APLOG_CRIT,rv,s,
                   "mod_spin: cannot create thread private key");
      return;
    }

    /* create thread mutex */
    if((rv=apr_thread_mutex_create(&mutex,APR_THREAD_MUTEX_DEFAULT,
                                   ppool))!=APR_SUCCESS){
      ap_log_error(APLOG_MARK,APLOG_CRIT,rv,s,
                   "mod_spin: cannot create mutex");
      return;
    }
  }
#endif

  /* initialise apr_dbd */
  if((rv=apr_dbd_init(ppool))!=APR_SUCCESS){
    ap_log_error(APLOG_MARK,APLOG_CRIT,rv,s,
                 "mod_spin: cannot initialise APR DBD");
    return;
  }

  /* regular expression for checking the validity of session cookies */
  ureg=ap_pregcomp(ppool,
                   apr_psprintf(ppool,"^[a-p]{%d}$",APR_MD5_DIGESTSIZE*4),
                   AP_REG_EXTENDED|AP_REG_NOSUB);

  /* check version and initialise XML parser */
  LIBXML_TEST_VERSION
}

/*
 * Configuration creation functions
 */
static rxv_spin_private_t *priv_config(request_rec *r){
  conn_rec *c=r->connection;
  rxv_spin_private_t *p;
  apr_status_t rv;

#if APR_HAS_THREADS
  /* get thread private data */
  if(mutex){
    if((rv=apr_threadkey_private_get((void *)(&p),tkey))!=APR_SUCCESS){
      ap_log_error(APLOG_MARK,APLOG_CRIT,rv,c->base_server,
                   "mod_spin: cannot get thread private data");
      return NULL;
    }
  } else{ /* no threads in MPM, use static variable */
    p=priv;
  }
#else
  p=priv;
#endif

  /* thread private data missing, create and store it for this thread */
  if(!p){
    p=private_create();
#if APR_HAS_THREADS
    if(mutex){
      /* set thread private data */
      if(apr_threadkey_private_set(p,tkey)!=APR_SUCCESS){
        apr_pool_destroy(p->pool);
        ap_log_error(APLOG_MARK,APLOG_CRIT,rv,c->base_server,
                     "mod_spin: cannot set thread private data");
        return NULL;
      }
    } else{ /* no threads in MPM, set static variable */
      priv=p;
    }
#else
   priv=p;
#endif
  }

  /* remember latest access time for connection reaping */
  if(r->request_time>p->atime)
    p->atime=r->request_time;

  return p;
}

static rxv_spin_reqcf_t *req_config(request_rec *r){
  rxv_spin_reqcf_t *rcnf;

  rcnf=apr_pcalloc(r->pool,sizeof(*rcnf));
  ap_set_module_config(r->request_config,&spin_module,rcnf);

  return rcnf;
}

static int app_config(request_rec *r,rxv_spin_config_t *conf,
                      rxv_spin_reqcf_t *rcnf,rxv_spin_private_t *priv,
                      rxv_spin_ctx_t *ctx){
  apr_status_t rv;
  apr_dso_handle_t *handle=NULL;
  apr_dso_handle_sym_t init=NULL;

  /* load the DSO, either from cache or new */
  if((rv=rxv_spin_dso_load(ctx,conf->application,&handle))!=APR_SUCCESS){
    if(handle){
      char *error=apr_palloc(r->pool,ERROR_BUF_SIZE);

      apr_dso_error(handle,error,ERROR_BUF_SIZE);
      ap_log_rerror(APLOG_MARK,APLOG_ERR,rv,r,"mod_spin: %s",error);
    } else{
      ap_log_rerror(APLOG_MARK,APLOG_ERR,rv,r,
                    "mod_spin: cannot load %s",conf->application);
    }

    return HTTP_INTERNAL_SERVER_ERROR;
  }

  if((rv=apr_dso_sym(&rcnf->svc,handle,conf->appentry)!=APR_SUCCESS)){
    char *error=apr_palloc(r->pool,ERROR_BUF_SIZE);

    apr_dso_error(handle,error,ERROR_BUF_SIZE);
    ap_log_rerror(APLOG_MARK,APLOG_ERR,rv,r,"mod_spin: %s",error);

    return HTTP_INTERNAL_SERVER_ERROR;
  }

  /* we don't care if fixup doesn't exist, so no error checking */
  apr_dso_sym(&rcnf->fix,handle,conf->appfixup);

  /* initialise the app */
  if(apr_dso_sym(&init,handle,conf->appinit)==APR_SUCCESS){
#if APR_HAS_THREADS
    apr_thread_mutex_lock(mutex);
#endif
    (*((rxv_spin_init_t)init))(ctx);
#if APR_HAS_THREADS
    apr_thread_mutex_unlock(mutex);
#endif
  }

  /* put context into the request configuration */
  rcnf->ctx=ctx;

  return OK;
}

/*
 * The fixup hook...
 */
static int fixup(request_rec *r){
  int rc;
  rxv_spin_config_t *conf;
  rxv_spin_reqcf_t *rcnf;
  rxv_spin_ctx_t *ctx;
  rxv_spin_private_t *priv;

  /* if the configuration is missing, big trouble */
  if(!(conf=ap_get_module_config(r->per_dir_config,&spin_module))){
    ap_log_rerror(APLOG_MARK,APLOG_ERR,APR_EGENERAL,r,
                  "mod_spin: missing configuration data");
    return HTTP_INTERNAL_SERVER_ERROR;
  }

  /* get private configuration, or create it if not there */
  if(!(priv=priv_config(r)))
    return HTTP_INTERNAL_SERVER_ERROR;

  /* make sure connection pool is cleaned up */
  apr_pool_cleanup_register(r->pool,priv,private_cleanup,apr_pool_cleanup_null);

  /* none of our business, don't handle */
  if(!(conf->application || conf->store || conf->cookie))
    return DECLINED;

  /* get request configuration, or create it if not there */
  if(!(rcnf=ap_get_module_config(r->request_config,&spin_module)) &&
     !(rcnf=req_config(r)))
    return HTTP_INTERNAL_SERVER_ERROR;

  /* prepare session identifier if configured */
  if(conf->cookie && salt){
    apreq_handle_t *req=apreq_handle_apache2(r);

    /* get/set session identifier and serve it back to the client */
    if(req){
      char *sesck=NULL,*uniq,*hash;
      apreq_cookie_t *cookie;

      /* we have session ID already */
      if((cookie=apreq_jar_get(req,conf->cookie))){
        /* extract session from he cookie */
        if(*cookie->v.data=='"' && *(cookie->v.data+cookie->v.dlen-1)=='"')
          sesck=apr_pstrmemdup(r->pool,cookie->v.data+1,cookie->v.dlen-2);
        else
          sesck=apr_pstrmemdup(r->pool,cookie->v.data,cookie->v.dlen);

        /* verify session ID */
        if(!sesck || ap_regexec(ureg,sesck,0,NULL,0)){
          ap_log_rerror(APLOG_MARK,APLOG_ERR,APR_EACCES,r,
                        "mod_spin: invalid session identifier");
          return HTTP_INTERNAL_SERVER_ERROR;
        }

        /* unique id will be session identifier, hash is the rest */
        hash=sesck+APR_MD5_DIGESTSIZE*2;
        uniq=apr_pstrmemdup(r->pool,sesck,APR_MD5_DIGESTSIZE*2);

        /* do crypto validation against new and possibly old salt */
        if(sesck_valid(r->pool,uniq,hash)!=APR_SUCCESS){
          ap_log_rerror(APLOG_MARK,APLOG_ERR,APR_EACCES,r,
                        "mod_spin: invalid session identifier");
          return HTTP_INTERNAL_SERVER_ERROR;
        }

        /* all clear, we have the session and it is valid */
        rcnf->sesid=uniq;
        rcnf->valid=1;
      } else{ /* we have to build session ID ourselves */
        if((uniq=(char *)apr_table_get(r->subprocess_env,"UNIQUE_ID"))){
          if(!((uniq=rxv_spin_hash(r->pool,uniq)) &&
               (hash=rxv_spin_hmac(r->pool,uniq,salt)))){
            ap_log_rerror(APLOG_MARK,APLOG_ERR,APR_ENOMEM,r,
                          "mod_spin: hashing failed");
            return HTTP_INTERNAL_SERVER_ERROR;
          }

          sesck=apr_pstrcat(r->pool,uniq,hash,NULL);
        }

        /* set session, but not valid yet */
        rcnf->sesid=uniq;
      }

      /* make and bake the cookie */
      if(sesck){
        char *scookie;

        if(!(cookie=apreq_cookie_make(r->pool,
                                      conf->cookie,strlen(conf->cookie),
                                      sesck,APR_MD5_DIGESTSIZE*4))){
          ap_log_rerror(APLOG_MARK,APLOG_ERR,APR_ENOMEM,r,
                        "mod_spin: cannot make session cookie");
          return HTTP_INTERNAL_SERVER_ERROR;
        }
        cookie->path="/";

        if(!(scookie=apreq_cookie_as_string(cookie,r->pool))){
          ap_log_rerror(APLOG_MARK,APLOG_ERR,APR_ENOMEM,r,
                        "mod_spin: cannot serialise session cookie");
          return HTTP_INTERNAL_SERVER_ERROR;
        }
        apr_table_addn(r->err_headers_out,"Set-Cookie",scookie);
      }
    }
  }

  /* stop processing here for sub-requests */
  if(r->main)
    return DECLINED;

  /* we have to have minimum configuration to do anything */
  if(conf->store || conf->application){
    ctx=context_create(r->pool);

    /* connection pooling */
    ctx->cpool=conf->connpool;

    /* raw request, private data, configuration file and session */
    ctx->r=r;
    ctx->priv=priv;
    ctx->cfgfn=conf->cfgfn;
    ctx->havest=(conf->driver?RXV_SPIN_TRUE:RXV_SPIN_FALSE);
    ctx->dbstor=(ctx->havest?(strcmp(conf->driver,"file")?RXV_SPIN_TRUE:
                                                          RXV_SPIN_FALSE):
                             RXV_SPIN_FALSE);
    ctx->cinfo=conf->store;
    ctx->appfn=conf->appfn;
    ctx->sttbl=conf->storetbl;
    ctx->sesid=rcnf->sesid;
    ctx->valid=rcnf->valid;
    ctx->atime=r->request_time;
    ctx->tmout=conf->timeout;

    /* find and execute prepare function */
    if(conf->application){
      if((rc=app_config(r,conf,rcnf,priv,ctx))!=OK)
        return rc;

      /* call prepare function only if we have one */
      if(rcnf->fix){
        rc=(*((rxv_spin_prepare_t)rcnf->fix))(ctx);

        if(ap_is_HTTP_ERROR(rc)){
          ctx->havest=RXV_SPIN_FALSE; /* don't save to store on error */
          ap_log_rerror(APLOG_MARK,APLOG_ERR,APR_EGENERAL,r,
                        "mod_spin: application error in prepare");
          return rc;
        }

        return rc;
      }
    }
  }

  return OK;
}

/*
 * The handlers...
 */
static int handler(request_rec *r){
  int rc;
  apr_status_t rv;
  rxv_spin_config_t *conf;
  rxv_spin_reqcf_t *rcnf;
  rxv_spin_extra_t *extra;
  rxv_spin_ctx_t *ctx;
  rxv_spin_private_t *priv;
  apr_file_t *fp;
  apr_finfo_t *fi;
  apr_mmap_t *mm;
  apr_bucket *b;

  /* don't do what isn't ours */
  if(strcmp(r->handler,"spin-template") && strcmp(r->handler,SPIN_MAGIC_TYPE))
    return DECLINED;

  /* if the configuration is missing, big trouble */
  if(!(conf=ap_get_module_config(r->per_dir_config,&spin_module))){
    ap_log_rerror(APLOG_MARK,APLOG_ERR,APR_EGENERAL,r,
                  "mod_spin: missing configuration data");
    return HTTP_INTERNAL_SERVER_ERROR;
  }

  /* get private and request configuration, or create them if not there */
  if(!(priv=priv_config(r)))
    return HTTP_INTERNAL_SERVER_ERROR;
  if(!(rcnf=ap_get_module_config(r->request_config,&spin_module)) &&
     !(rcnf=req_config(r)))
    return HTTP_INTERNAL_SERVER_ERROR;

  /* build the necessary memory structures */
  extra=extra_create(r,rcnf->ctx);

  /* some important shortcuts before anything else */
  ctx=extra->ctx;

  /* if we didn't populate context during the fixup phase, do it now */
  if(!rcnf->ctx){
    /* connection pooling */
    ctx->cpool=conf->connpool;

    /* raw request, configuration file and session */
    ctx->r=r;
    ctx->priv=priv;
    ctx->cfgfn=conf->cfgfn;
    ctx->havest=(conf->driver?RXV_SPIN_TRUE:RXV_SPIN_FALSE);
    ctx->dbstor=(ctx->havest?(strcmp(conf->driver,"file")?RXV_SPIN_TRUE:
                                                          RXV_SPIN_FALSE):
                             RXV_SPIN_FALSE);
    ctx->cinfo=conf->store;
    ctx->appfn=conf->appfn;
    ctx->sttbl=conf->storetbl;
    ctx->sesid=rcnf->sesid;
    ctx->valid=rcnf->valid;
    ctx->atime=r->request_time;
    ctx->tmout=conf->timeout;
  }

  /* grab the apreq handle for cookie and parameter parsing */
  ctx->req=apreq_handle_apache2(r);

  /* run input filters and discard the body, all access via libapreq2 */
  if((rc=ap_discard_request_body(r))!=OK){
    ap_log_rerror(APLOG_MARK,APLOG_ERR,APR_EGENERAL,r,
                  "mod_spin: cannot run input filters");
    return rc;
  }

  /* find and execute application */
  if(conf->application){
    if(!rcnf->ctx && (rc=app_config(r,conf,rcnf,priv,ctx))!=OK)
      return rc;

    /* call entry function */
    rc=(*((rxv_spin_service_t)rcnf->svc))(ctx);

    if(ap_is_HTTP_ERROR(rc)){
      ctx->havest=RXV_SPIN_FALSE; /* don't save to store on error */
      ap_log_rerror(APLOG_MARK,APLOG_ERR,APR_EGENERAL,r,
                    "mod_spin: application error in service");
      return rc;
    }

    /* not a failure, but don't process the template */
    if(rc!=OK)
      return rc;
  } else{
    ap_log_rerror(APLOG_MARK,APLOG_INFO,APR_SUCCESS,r,
                  "mod_spin: application not specified");
  }

  /* check existence of the template */
  if(!r->finfo.filetype){
    ap_log_rerror(APLOG_MARK,APLOG_ERR,APR_ENOENT,r,
                  "mod_spin: file does not exist: %s",r->filename);
    return HTTP_NOT_FOUND;
  }

  /* always open file - we want to be atomic at inode level */
  if((rv=apr_file_open(&fp,r->filename,APR_FOPEN_READ,
                       APR_FPROT_OS_DEFAULT,r->pool))!=APR_SUCCESS){
    ap_log_rerror(APLOG_MARK,APLOG_ERR,rv,r,
                  "mod_spin: cannot open file: %s",r->filename);
    return HTTP_FORBIDDEN;
  }

  /* even if the file is fully cached, reparse if changed */
  fi=apr_pcalloc(r->pool,sizeof(*fi));
  if((rv=apr_file_info_get(fi,APR_FINFO_SIZE,fp))!=APR_SUCCESS){
    ap_log_rerror(APLOG_MARK,APLOG_ERR,rv,r,
                  "mod_spin: cannot stat file: %s",r->filename);
    return HTTP_FORBIDDEN;
  }

  /* mmap the file */
  if((rv=apr_mmap_create(&mm,fp,
                         0,fi->size,APR_MMAP_READ,r->pool))!=APR_SUCCESS){
    ap_log_rerror(APLOG_MARK,APLOG_ERR,rv,r,
                  "mod_spin: cannot mmap file: %s",r->filename);
    return HTTP_INTERNAL_SERVER_ERROR;
  }

  /* close file to save file descriptors */
  apr_file_close(fp);

  /* mmap bucket that we'll share around */
  if(!(b=apr_bucket_mmap_create(mm,0,fi->size,extra->output->bucket_alloc))){
    ap_log_rerror(APLOG_MARK,APLOG_ERR,APR_ENOMEM,r,
                  "mod_spin: cannot create mmap bucket");
    return HTTP_INTERNAL_SERVER_ERROR;
  }

  /* create brigade for scanning */
  extra->input=apr_brigade_create(r->pool,extra->output->bucket_alloc);
  APR_BRIGADE_INSERT_HEAD(extra->input,b);
  APR_BUCKET_INSERT_AFTER(b,apr_bucket_eos_create(extra->output->bucket_alloc));

  while(extra->input){
    /* reset root, count of buckets to queue and clear node pool */
    extra->root=NULL;
    extra->count=0;
    apr_pool_clear(extra->pool);

    if((rv=rxv_spin_parse(extra))!=APR_SUCCESS){
      if(rv==APR_EACCES){
        ap_log_rerror(APLOG_MARK,APLOG_ERR,rv,r,
                      "mod_spin: permission denied: %s",r->filename);
        return HTTP_FORBIDDEN;
      } else{
        ap_log_rerror(APLOG_MARK,APLOG_ERR,rv,r,
                      "mod_spin: parsing error: %s: %s",
                      r->filename,extra->error?extra->error:"unspecified");
        return HTTP_INTERNAL_SERVER_ERROR;
      }
    }

    /* finally, process the template */
    if(extra->root){
      if((rv=rxv_spin_render(extra->root,extra))!=APR_SUCCESS){
        ap_log_rerror(APLOG_MARK,APLOG_ERR,rv,r,
                      "mod_spin: template processing error: %s: %s",
                      r->filename,extra->error?extra->error:"unspecified");
        return HTTP_INTERNAL_SERVER_ERROR;
      }

      /* very end, add EOS bucket */
      if(extra->eos && !extra->input){
        APR_BUCKET_REMOVE(extra->eos);
        APR_BRIGADE_INSERT_TAIL(extra->output,extra->eos);
      }

      if(!APR_BRIGADE_EMPTY(extra->output) &&
         (rv=ap_pass_brigade(r->output_filters,extra->output))!=APR_SUCCESS){
        ap_log_rerror(APLOG_MARK,APLOG_ERR,rv,r,
                      "mod_spin: cannot pass brigade to filters");
        return HTTP_INTERNAL_SERVER_ERROR;
      }
    }

    /* free up some space */
    apr_brigade_cleanup(extra->output);
    apr_brigade_cleanup(extra->text);
  }

  return OK;
}

/*
 * The filters...
 */
static apr_status_t filter(ap_filter_t *f,apr_bucket_brigade *b){
  apr_status_t rv;
  request_rec *r=f->r;
  rxv_spin_fltcf_t *fcnf=f->ctx;
  rxv_spin_extra_t *extra;

  /* fast exit */
  if(APR_BRIGADE_EMPTY(b))
    return APR_SUCCESS;

  /* been here before, just pick up existing stuff */
  if(fcnf){
    extra=fcnf->extra;

    /* reset positions, EOF flag, error etc. */
    extra->offset=0;
    extra->pos=-1;
    extra->soff=0;
    extra->blen=extra->bpos=0;
    extra->eof=NULL;
    extra->error=NULL;
    extra->split=NULL;
    extra->depth=extra->ldepth=0;

    /* initialise parsing variables */
    memset(extra->tail,0,sizeof(*extra->tail)*RXV_SPIN_MAX_DEPTH*2);
    memset(extra->loop,0,sizeof(*extra->loop)*RXV_SPIN_MAX_DEPTH);
    memset(extra->nest,0,sizeof(*extra->nest)*RXV_SPIN_MAX_DEPTH);
    memset(extra->boff,0,sizeof(*extra->boff)*RXV_SPIN_MAX_DEPTH);
  } else{ /* first time through, create filter config */
    int rc;
    rxv_spin_config_t *conf;
    rxv_spin_reqcf_t *rcnf;
    rxv_spin_ctx_t *ctx;
    rxv_spin_private_t *priv;

    /* new filter context */
    f->ctx=fcnf=apr_pcalloc(r->pool,sizeof(*fcnf));

    /* if the configuration is missing, big trouble */
    if(!(conf=ap_get_module_config(r->per_dir_config,&spin_module))){
      ap_log_rerror(APLOG_MARK,APLOG_ERR,APR_EGENERAL,r,
                    "mod_spin: missing configuration data");
      ap_remove_output_filter(f);
      return ap_pass_brigade(f->next,b);
    }

    /* get private and request configuration, or create them if not there */
    if(!(priv=priv_config(r))){
      ap_remove_output_filter(f);
      return ap_pass_brigade(f->next,b);
    }
    if(!(rcnf=ap_get_module_config(r->request_config,&spin_module)) &&
       !(rcnf=req_config(r))){
      ap_remove_output_filter(f);
      return ap_pass_brigade(f->next,b);
    }

    /* build the necessary memory structures */
    fcnf->extra=extra=extra_create(r,rcnf->ctx);

    /* some important shortcuts before anything else */
    ctx=extra->ctx;

    /* if we didn't populate context during the fixup phase, do it now */
    if(!rcnf->ctx){
      /* connection pooling */
      ctx->cpool=conf->connpool;

      /* raw request, configuration file and session */
      ctx->r=r;
      ctx->priv=priv;
      ctx->cfgfn=conf->cfgfn;
      ctx->havest=(conf->driver?RXV_SPIN_TRUE:RXV_SPIN_FALSE);
      ctx->dbstor=(ctx->havest?(strcmp(conf->driver,"file")?RXV_SPIN_TRUE:
                                                            RXV_SPIN_FALSE):
                               RXV_SPIN_FALSE);
      ctx->cinfo=conf->store;
      ctx->appfn=conf->appfn;
      ctx->sttbl=conf->storetbl;
      ctx->sesid=rcnf->sesid;
      ctx->valid=rcnf->valid;
      ctx->atime=r->request_time;
      ctx->tmout=conf->timeout;
    }

    /* grab the apreq handle for cookie and parameter parsing */
    ctx->req=apreq_handle_apache2(r);

    /* run input filters and discard the body, all access via libapreq2 */
    if((rc=ap_discard_request_body(r))!=OK){
      ap_log_rerror(APLOG_MARK,APLOG_ERR,APR_EGENERAL,r,
                    "mod_spin: cannot run input filters");
      ap_remove_output_filter(f);
      return ap_pass_brigade(f->next,b);
    }

    /* find and execute application */
    if(conf->application){
      if(!rcnf->ctx && (rc=app_config(r,conf,rcnf,priv,ctx))!=OK){
        ap_remove_output_filter(f);
        return ap_pass_brigade(f->next,b);
      }

      /* call entry function */
      rc=(*((rxv_spin_service_t)rcnf->svc))(ctx);

      /* anything but OK is an error here */
      if(rc!=OK){
        ctx->havest=RXV_SPIN_FALSE; /* don't save to store on error */
        ap_log_rerror(APLOG_MARK,APLOG_ERR,APR_EGENERAL,r,
                      ap_is_HTTP_ERROR(rc)?
                        "mod_spin: application error in service":
                        "mod_spin: application returned an invalid code");
        ap_remove_output_filter(f);
        return ap_pass_brigade(f->next,b);
      }
    } else{
      ap_log_rerror(APLOG_MARK,APLOG_INFO,APR_SUCCESS,r,
                    "mod_spin: application not specified");
    }

    /* filter specific knickknacks */
    extra->f=f;
    extra->metabb=apr_brigade_create(r->pool,r->connection->bucket_alloc);
  }

  /* concatenate saved and input brigade */
  if(extra->input)
    APR_BRIGADE_CONCAT(extra->input,b);
  else
    extra->input=b;

  /* we are going to modify content, so this must be unset */
  apr_table_unset(r->headers_out,"Content-Length");
  apr_table_unset(r->headers_out,"Last-Modified");

  while(extra->input){
    /* reset root, count of buckets to queue and clear node pool */
    extra->root=NULL;
    extra->count=0;
    apr_pool_clear(extra->pool);

    if((rv=rxv_spin_parse(extra))!=APR_SUCCESS){
      if(rv==APR_EACCES)
        ap_log_rerror(APLOG_MARK,APLOG_ERR,rv,r,
                      "mod_spin: permission denied: %s",r->filename);
      else
        ap_log_rerror(APLOG_MARK,APLOG_ERR,rv,r,
                      "mod_spin: parsing error: %s: %s",
                      r->filename,extra->error?extra->error:"unspecified");

      ap_remove_output_filter(f);
      return ap_pass_brigade(f->next,b);
    }

    /* finally, process the template */
    if(extra->root){
      if((rv=rxv_spin_render(extra->root,extra))!=APR_SUCCESS){
        ap_log_rerror(APLOG_MARK,APLOG_ERR,rv,r,
                      "mod_spin: template processing error: %s: %s",
                      r->filename,extra->error?extra->error:"unspecified");
        ap_remove_output_filter(f);
        return ap_pass_brigade(f->next,b);
      }

      /* very end, add EOS bucket */
      if(extra->eos && !extra->input){
        APR_BUCKET_REMOVE(extra->eos);
        APR_BRIGADE_INSERT_TAIL(extra->metabb,extra->eos);
      }

      /* glue on metadata */
      APR_BRIGADE_CONCAT(extra->output,extra->metabb);

      if(!APR_BRIGADE_EMPTY(extra->output) &&
         (rv=ap_pass_brigade(f->next,extra->output))!=APR_SUCCESS){
        ap_log_rerror(APLOG_MARK,APLOG_ERR,rv,r,
                      "mod_spin: cannot pass brigade to filters");
        ap_remove_output_filter(f);
        return ap_pass_brigade(f->next,b);
      }
    }

    /* free up some space */
    apr_brigade_cleanup(extra->output);
    apr_brigade_cleanup(extra->text);
  }

  /* setaside split brigade, if any */
  if(extra->split)
    ap_save_brigade(f,&extra->input,&extra->split,r->pool);

  /* get rid of the original brigade */
  apr_brigade_cleanup(b);

  return APR_SUCCESS;
}

/*
 * Directives...
 */
static const command_rec spin_commands[]={
  AP_INIT_TAKE1("SpinApplication",set_application,NULL,ACCESS_CONF|RSRC_CONF,
                "path to the shared library of the application"),
  AP_INIT_TAKE1("SpinAppInit",set_appinit,NULL,ACCESS_CONF|RSRC_CONF,
                "name of the init function in the shared library"),
  AP_INIT_TAKE1("SpinAppPrepare",set_appfixup,NULL,ACCESS_CONF|RSRC_CONF,
                "name of the prepare function in the shared library"),
  AP_INIT_TAKE1("SpinAppService",set_appentry,NULL,ACCESS_CONF|RSRC_CONF,
                "name of the service function in the shared library"),
  AP_INIT_TAKE1("SpinAppConfig",set_appconfig,NULL,ACCESS_CONF|RSRC_CONF,
                "path to the configuration file"),
  AP_INIT_TAKE1("SpinStore",set_store,NULL,ACCESS_CONF|RSRC_CONF,
                "store connect string"),
  AP_INIT_TAKE1("SpinStoreTable",set_storetbl,NULL,ACCESS_CONF|RSRC_CONF,
                "store table name"),
  AP_INIT_TAKE1("SpinCookie",set_cookie,NULL,ACCESS_CONF|RSRC_CONF,
                "name of the cookie for session tracking"),
  AP_INIT_TAKE1("SpinTimeout",set_timeout,NULL,ACCESS_CONF|RSRC_CONF,
                "session timeout in seconds"),
  AP_INIT_TAKE1("SpinConnPool",set_connpool,NULL,ACCESS_CONF|RSRC_CONF,
                "should connections be pooled"),
  AP_INIT_TAKE1("SpinConnCount",set_conncount,NULL,RSRC_CONF,
                "maximum number of connections in the pool"),
  AP_INIT_TAKE1("SpinSalt",set_salt,NULL,RSRC_CONF,
                "cryptographic salt"),
  {NULL}
};

static void register_hooks(apr_pool_t *p){
  ap_hook_post_config(post_config,NULL,NULL,APR_HOOK_MIDDLE);
  ap_hook_child_init(child_init,NULL,NULL,APR_HOOK_MIDDLE);
  ap_hook_fixups(fixup,NULL,NULL,APR_HOOK_MIDDLE);
  ap_hook_handler(handler,NULL,NULL,APR_HOOK_MIDDLE);
  ap_register_output_filter("SPIN-TEMPLATE",filter,NULL,AP_FTYPE_RESOURCE);
}

module AP_MODULE_DECLARE_DATA spin_module={
  STANDARD20_MODULE_STUFF,
  create_config_directory,    /* create per-dir config */
  merge_config,               /* merge per-dir config */
  create_config_server,       /* server config */
  merge_config,               /* merge server config */
  spin_commands,              /* command apr_table_t */
  register_hooks              /* register hooks */
};
