/*
 * private.h
 *
 * Copyright (C) 2003 - 2007 Bojan Smojver, Rexursive
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef __RXV_SPIN_PRIVATE__
#define __RXV_SPIN_PRIVATE__

#include <config.h>
#include <rxv_spin.h>

#if defined (HAVE_LIBPQ) && defined (HAVE_LIBPQ_FE_H)
#include <libpq-fe.h>
#endif
#if defined (HAVE_LIBMYSQLCLIENT_R) && defined (HAVE_MYSQL_H)
#include <mysql.h>
#endif
#if defined (HAVE_LIBSQLITE) && defined (HAVE_SQLITE_H)
#include <sqlite.h>
#endif
#if defined (HAVE_LIBSQLITE3) && defined (HAVE_SQLITE3_H)
#include <sqlite3.h>
#endif

/*
 * Type definitions
 */
typedef struct rxv_spin_node    rxv_spin_node_t;
typedef struct rxv_spin_txt     rxv_spin_txt_t;
typedef struct rxv_spin_ref     rxv_spin_ref_t;
typedef struct rxv_spin_for     rxv_spin_for_t;
typedef struct rxv_spin_if      rxv_spin_if_t;
typedef struct rxv_spin_reg     rxv_spin_reg_t;
typedef struct rxv_spin_lit     rxv_spin_lit_t;
typedef struct rxv_spin_num     rxv_spin_num_t;
typedef struct rxv_spin_extra   rxv_spin_extra_t;
typedef struct rxv_spin_config  rxv_spin_config_t;
typedef struct rxv_spin_conncf  rxv_spin_conncf_t;
typedef struct rxv_spin_reqcf   rxv_spin_reqcf_t;
typedef struct rxv_spin_fltcf   rxv_spin_fltcf_t;
typedef struct rxv_spin_private rxv_spin_private_t;
typedef struct rxv_spin_db_stmt rxv_spin_db_stmt_t;

/*
 * Generic true and false
 */
typedef enum{
  RXV_SPIN_FALSE,
  RXV_SPIN_TRUE
} rxv_spin_bool_e;

/*
 * Data and context structures
 */
struct rxv_spin_ctx{
  apr_pool_t         *pool;    /* context specific pool */
  apr_pool_t         *ppool;   /* process specific pool */
  rxv_spin_data_t    *data;    /* application prepared data */
  request_rec        *r;       /* raw Apache request */
  apreq_handle_t     *req;     /* parsed request */
  rxv_spin_private_t *priv;    /* thread private data */
  char               *cfgfn;   /* configuration file filename */
  char               *cinfo;   /* store connection string */
  char               *sttbl;   /* store table name */
  rxv_spin_db_t      *store;   /* store database connection */
  char               *appfn;   /* aplication store file name */
  char               *sesfn;   /* session store file name */
  char               *sesid;   /* session id */
  rxv_spin_bool_e    cpool:1;  /* connections are pooled */
  rxv_spin_bool_e    valid:1;  /* session is valid */
  rxv_spin_bool_e    sdirty:1; /* session store dirty */
  rxv_spin_bool_e    adirty:1; /* application store dirty */
  rxv_spin_bool_e    havest:1; /* we have the store */
  rxv_spin_bool_e    dbstor:1; /* store has database backend */
  rxv_spin_bool_e    closed:1; /* store still closed */
  apr_hash_t         *app;     /* application store */
  apr_hash_t         *ses;     /* session store */
  apr_time_t         atime;    /* time session was last accessed */
  apr_time_t         ptime;    /* time configuration was last parsed */
  apr_time_t         tmout;    /* session timeout in microseconds */
  const char         *sstore;  /* select timestamp and data from store */
  const char         *ustore;  /* update timestamp and data in store */
  const char         *istore;  /* insert timestamp and data into store */
  const char         *ustamp;  /* update timestamp in store */
  void               *extra;   /* extra data placed in context by API user */
};

/*
 * Database structures
 */
struct rxv_spin_db{ 
  APR_RING_ENTRY(rxv_spin_db) link;    /* LRU: ring link */
  apr_pool_t                  *pool;   /* connection specific pool */
  apr_pool_t                  *spool;  /* select (temporary) pool */
  char                        *cinfo;  /* identification: type:connect string */
  const apr_dbd_driver_t      *driver; /* APR Util database driver */
  apr_dbd_t                   *handle; /* APR Util database handle */
  rxv_spin_private_t          *priv;   /* thread private data */
  apr_status_t                (*cleanup)(void *data); /* cleanup function */
  apr_hash_t                  *stmts;  /* prepared statements */
  apr_time_t                  atime;   /* time connection was last accessed */
};

struct rxv_spin_db_txn{ 
  apr_pool_t            *pool; /* transaction specific pool */
  rxv_spin_db_t         *db;   /* database assiciated with this transaction */
  apr_dbd_transaction_t *txn;  /* APR Util transaction */
};

/*
 * Statement/node types
 */
typedef enum{
  RXV_SPIN_UDF,
  RXV_SPIN_TXT,
  RXV_SPIN_REF,
  RXV_SPIN_FOR,
  RXV_SPIN_IF,
  RXV_SPIN_REG,
  RXV_SPIN_LIT,
  RXV_SPIN_NUM
} rxv_spin_node_e;

/*
 * Reference types
 */
typedef enum{
  RXV_SPIN_REF_REG,
  RXV_SPIN_REF_SIZ,
  RXV_SPIN_REF_IND
} rxv_spin_ref_e;

/*
 * Conditional types: #if (guaranteed to be 0) and #unless (guaranteed to be 1)
 */
typedef enum{
  RXV_SPIN_IF_IF,
  RXV_SPIN_IF_UNLESS
} rxv_spin_if_e;

/*
 * Comparison operators
 */
typedef enum{
  RXV_SPIN_OP_NONE,
  RXV_SPIN_OP_MATCH,
  RXV_SPIN_OP_EQ_LIT,
  RXV_SPIN_OP_EQ_NUM,
  RXV_SPIN_OP_EQ_REF,
  RXV_SPIN_OP_MODULO
} rxv_spin_op_e;

/*
 * Maximum nesting depth of the #if/#unless and #for
 */
#define RXV_SPIN_MAX_DEPTH 32

/*
 * Maximum/minimum size of scan buffer
 */
#define RXV_SPIN_SCAN_MIN APR_BUCKET_BUFF_SIZE
#define RXV_SPIN_SCAN_MAX (RXV_SPIN_SCAN_MIN*8)

/*
 * Maximum number of buckets to hold before we flush
 */
#define RXV_SPIN_FLUSH_COUNT 64

/*
 * Data structures for Abstract Syntax Tree
 */
struct rxv_spin_node{
  rxv_spin_node_e type:3;
  rxv_spin_node_t *next;
};

struct rxv_spin_txt{
  rxv_spin_node_t node;
  apr_bucket      *bucket; /* ready to use/copy bucket */
  rxv_spin_bool_e used:1;  /* have we used this bucket already */
};

struct rxv_spin_ref{
  rxv_spin_node_t node;
  char            *name;    /* full name of the reference */
  char            *chunk;   /* last chunk of the full name */
  rxv_spin_data_t *data;    /* data currently associated with this reference */
  rxv_spin_for_t  *loop;    /* points to its for loop, if any */
  rxv_spin_bool_e looked:1; /* did we look for the data before */
  rxv_spin_ref_e  type:2;   /* type of reference: regular, size or index */
};

struct rxv_spin_for{
  rxv_spin_node_t node;
  rxv_spin_ref_t  *ref;  /* what to loop around */
  rxv_spin_node_t *stmt; /* what to replicate each time */
  size_t          index; /* the index of the loop */
};

struct rxv_spin_if{
  rxv_spin_node_t node;
  rxv_spin_if_e  type:1; /* is this actually the unless, not if */
  rxv_spin_op_e  oper:3; /* what operator is used */
  rxv_spin_ref_t *ref;   /* the reference we're checking, on the left */
  union{
    ap_regex_t     *reg; /* regular expression to match */
    rxv_spin_ref_t *rfr; /* the reference to match, on the right */
    char           *lit; /* literal string */
    struct{
      unsigned long num; /* number to the right of the operator */
      unsigned long res; /* result of the modulo comparison, right of == */
    };
  };
  rxv_spin_node_t *pos;  /* statement if ref is not null */
  rxv_spin_node_t *neg;  /* statement if ref is null */
};

struct rxv_spin_reg{
  rxv_spin_node_t node;
  ap_regex_t      reg;
};

struct rxv_spin_lit{
  rxv_spin_node_t node;
  char            *lit;
};

struct rxv_spin_num{
  rxv_spin_node_t node;
  unsigned long   num;
};

/*
 * Data structures for the scanner
 */
struct rxv_spin_extra{
  apr_pool_t         *pool;      /* regular pool for all memory allocations */
  apr_off_t          offset;     /* from beginning of brigade */
  apr_off_t          pos;        /* good chunk from beginning of brigade */
  apr_off_t          row;        /* current row of the template */
  apr_off_t          col;        /* current column of the template */
  rxv_spin_node_t    *root;      /* root of the abstract syntax tree */
  rxv_spin_node_t    **tail;     /* tails of node lists on the stack */
  rxv_spin_for_t     **loop;     /* nesting loops */
  unsigned char      *nest;      /* nesting type stack, #if or #for */
  size_t             *boff;      /* branch offset, RXV_SPIN_MAX_DEPTH for #else */
  int                depth;      /* current nesting depth */
  int                ldepth;     /* current loop nesting depth */
  apr_size_t         count;      /* number of parsed TXT/REF tokens */
  char               *error;     /* descriptive error */
  apr_bucket_brigade *input;     /* what we're scanning */
  apr_bucket_brigade *split;     /* split brigade that needs saving */
  apr_bucket_brigade *output;    /* where the result goes */
  apr_bucket_brigade *metabb;    /* metadata bucket brigade to send out */
  apr_bucket_brigade *text;      /* brigade of text nodes */
  apr_bucket         *eof;       /* brigade sentinel seen */
  apr_bucket         *eos;       /* EOS bucket seen */
  apr_bucket         *scan;      /* first bucket we're scanning */
  apr_size_t         soff;       /* offset from start of the scan bucket */
  apr_off_t          erow;       /* scan row where the error occurred */
  apr_off_t          ecol;       /* scan column where the error occurred */
  char               *buf;       /* flex buffer */
  apr_size_t         blen;       /* flex buffer length */
  apr_size_t         bpos;       /* where are we within the buffer now */
  void               *scanner;   /* the scanner */
  rxv_spin_ctx_t     *ctx;       /* the context - data, parameters etc. */
  ap_filter_t        *f;         /* our filter, if any */
};

/* Function declarations */
apr_status_t rxv_spin_parse(rxv_spin_extra_t *extra);
apr_status_t rxv_spin_render(rxv_spin_node_t *node,rxv_spin_extra_t *extra);

/* Missing parser and scanner stuff */
#define YYSTYPE       rxv_spin_node_t*
#define YY_EXTRA_TYPE rxv_spin_extra_t*

int rxv_spin_yyparse(void *YYPARSE_PARAM);
void rxv_spin_yyerror(char *s);
int rxv_spin_yylex(rxv_spin_node_t **yyval_param,void *yyscanner);

/*
 * Module configuration
 */
typedef enum{
  RXV_SPIN_APPLICATION_SET=0x0001,
  RXV_SPIN_APPINIT_SET    =0x0002,
  RXV_SPIN_APPFIXUP_SET   =0x0004,
  RXV_SPIN_APPENTRY_SET   =0x0008,
  RXV_SPIN_CFGFN_SET      =0x0010,
  RXV_SPIN_STORE_SET      =0x0020,
  RXV_SPIN_STORETBL_SET   =0x0040,
  RXV_SPIN_COOKIE_SET     =0x0080,
  RXV_SPIN_TIMEOUT_SET    =0x0100,
  RXV_SPIN_CONNPOOL_SET   =0x0200
} rxv_spin_cfg_e;

struct rxv_spin_config{
  char            *application; /* path to the application shared library */
  char            *appinit;     /* application init function name */
  char            *appfixup;    /* application fixup function name */
  char            *appentry;    /* application entry function name */
  char            *cfgfn;       /* configuration file filename */
  char            *driver;      /* store connect driver */
  char            *store;       /* store connect string */
  char            *appfn;       /* application store file name */
  char            *storetbl;    /* store table name */
  char            *cookie;      /* name of the cookie used for session tracking */
  apr_time_t      timeout;      /* session timeout */
  rxv_spin_bool_e connpool:1;   /* enable connection pooling */
  rxv_spin_cfg_e  set:10;       /* what options were actually set */
};

/*
 * Request configuration
 */
struct rxv_spin_reqcf{
  char                 *sesid;  /* session id */
  rxv_spin_ctx_t       *ctx;    /* context passed from fixup to handler */
  apr_dso_handle_sym_t fix;     /* fixup function handle */
  apr_dso_handle_sym_t svc;     /* service function handle */
  rxv_spin_bool_e      valid:1; /* session is valid: value 1, otherwise 0 */
};

/*
 * Filter configuration
 */
struct rxv_spin_fltcf{
  rxv_spin_extra_t *extra;
};

/*
 * Private thread data
 */
struct rxv_spin_private{
  apr_pool_t *pool;  /* thread specific pool to avoid concurrency */
  apr_hash_t *dso;   /* cache of DSOs so we don't keep opening */
  apr_time_t atime;  /* time data was last accessed */
  apr_hash_t *conns; /* hash table of connections */
  unsigned   count;  /* connections in the pool */
  APR_RING_HEAD(rxv_spin_db_list,rxv_spin_db) list; /* LRU: conns list */
};

/*
 * Useful macros
 */
#define RXV_SPIN_UNS_DIGITS(t) ((sizeof(t)*5)/2+sizeof(t)%2)

/*
 * Crypto lengths
 */
#define RXV_SPIN_SALT_MIN 30
#define RXV_SPIN_SALT_MAX 64

#endif
